/* MIX simulator, copyright 1994 by Darius Bacon */ 
#include "mix.h"

#include <stdio.h>

Cell ulong2Cell(unsigned long n)
{
    if (CELL_MAX < n) {
	warn("Value out of range: %lu", n);
	return CELL_MAX;
    } else
	return n;
}

/* --- Field operations --- */

Byte get_byte(unsigned F, Cell cell)
{
    assert(F != 0);
    return (Byte) field(make_field_spec(F, F), cell);
}

Cell set_byte(Byte value, unsigned F, Cell into)
{
    assert(F != 0);
    return set_field((Cell) value, make_field_spec(F, F), into);
}

static Flag bad_field[64];  /* 64 = number of byte values */
static unsigned shift[64];
static long mask[64];

Cell field(Byte F, Cell cell)
{
    if (bad_field[F])
	error("Bad field spec: %02o", F);
    if (F < 8)      /* if F is of the form (0:R), retain the sign of -cell- */
	return ((cell & mask[F]) >> shift[F]) | sign_bit(cell);
    else
	return (cell & mask[F]) >> shift[F];
}

Cell set_field(Cell value, Byte F, Cell into)
{
    long m = mask[F];
    if (bad_field[F])
	error("Bad field spec: %02o", F);
    if (F < 8)      /* if F is of the form (0:R), use the sign of -value- */
	return (into & ~m & ~the_sign_bit) | ((value << shift[F]) & m) | sign_bit(value);
    else
	return (into & ~m) | ((value << shift[F]) & m);
}

void precompute_field_data(void)
{
    unsigned L, R;
    for (L = 0; L < 8; ++L)
	for (R = 0; R < 8; ++R) {
	    unsigned F = 8 * L + R;
	    bad_field[F] = R < L || 5 < R;
	    if (bad_field[F])
		shift[F] = 0, mask[F] = 0;
	    else {
		unsigned width = R - (L == 0 ? 1 : L) + 1;
		shift[F] = 6 * (5 - R);
		mask[F] = ((1L << (6 * width)) - 1) << shift[F];
	    } 
	}
}

Byte make_field_spec(unsigned L, unsigned R)
{
    unsigned F = 8 * L + R;
    assert(!bad_field[F]);
    return F;
}

void assert_valid_field(Cell field_spec)
{
    if (is_negative(field_spec) 
	 || 64 <= magnitude(field_spec)
	 || bad_field[(unsigned)magnitude(field_spec)]) {
	char buffer[12];
	unparse_cell(buffer, field_spec);
	error("Invalid field specifier: %s", buffer);
    }
}

/* --- Arithmetic --- */

Flag overflow = false;

Cell add(Cell x, Cell y)
{   /* This is kinda clumsy. Should I combine code at the cost (?) of speed */
    /* and functional style? */
    if (sign_bit(x) == sign_bit(y)) {
	long sum = magnitude(x) + magnitude(y);
	long magnitude_of_sum = magnitude(sum);
	if (magnitude_of_sum != sum) overflow = true;
	return sign_bit(x) | magnitude_of_sum;
    } else {
	long diff = magnitude(x) - magnitude(y);
	if (diff < 0) {
	    long sum = -diff;
	    long magnitude_of_sum = magnitude(sum);
	    assert(magnitude_of_sum == sum);    /* Overflow can't occur. */
	    return sign_bit(y) | magnitude_of_sum;
	} else {
	    long magnitude_of_diff = magnitude(diff);
	    assert(magnitude_of_diff == diff);
	    return sign_bit(x) | magnitude_of_diff;
	}
    }
}

Cell sub(Cell x, Cell y)
{
    return add(x, negative(y));     /* should inline this, maybe */
}

#define low16(cell)     ( (cell) & ((1L << 16) - 1) )   /* portable but slower than a cast */

void multiply(Cell x, Cell y, Cell *high_word, Cell *low_word)
{
    long sign = sign_bit(x) ^ sign_bit(y);
    unsigned long x0 = low16(x);
    unsigned long x1 = magnitude(x) >> 16;
    unsigned long y0 = low16(y);
    unsigned long y1 = magnitude(y) >> 16;
    unsigned long partials = x0 * y1 + x1 * y0;
    unsigned long low_sum = x0 * y0 + (low16(partials) << 16);
    unsigned long low_magnitude = magnitude(low_sum);
    *low_word = sign | low_magnitude;
    {
	unsigned long high_sum = 
	  x1 * y1 + (partials >> 16) + (low_magnitude != low_sum);
	unsigned long high_magnitude = magnitude(high_sum);
	assert(high_magnitude == high_sum);     /* overflow -should- be impossible, but I'll leave this in here for the moment */
	*high_word = sign | high_magnitude;
    }
}

Cell mul(Cell x, Cell y)
{
    Cell lo, hi;
    multiply(x, y, &hi, &lo);
    if (magnitude(hi) != 0) overflow = true;
    return lo;
}

void divide(Cell n1, Cell n0, Cell d, Cell *quotient, Cell *remainder)
{
    long magn1 = magnitude(n1);
    long magd = magnitude(d);
    if (magd == 0) {
	overflow = true;
	*quotient = *remainder = zero;  /* just so they have -some- valid value */
    } else if (magn1 == 0) {    /* special-cased for speed */
	*quotient = (sign_bit(n1) ^ sign_bit(d)) | (magnitude(n0) / magd);
	*remainder = sign_bit(n1) | (magnitude(n0) % magd);
    } else if (magd <= magn1) {
	overflow = true;
	*quotient = *remainder = zero;
    } else {
	long q = magnitude(n0);
	long r = magn1;
	unsigned i;
	for (i = 30; i != 0; --i) {
	    r <<= 1;
	    if (q & (1L << 30))
		++r;
	    q = (q << 1) & ((1L << 30) - 1);
	    if (magd <= r)
		++q, r -= magd;
	}
	*quotient = (sign_bit(n1) ^ sign_bit(d)) | q;
	*remainder = sign_bit(n1) | r;
    }
}

Cell slash(Cell x, Cell y)      /* the name 'div' is taken... */
{
    Cell quotient, remainder;
    divide(sign_bit(x), x, y, &quotient, &remainder);
    return quotient;
}

/* --- Shift operations --- */

void shift_right(Cell A, Cell X, unsigned long count, Cell *pA, Cell *pX)
{
    *pX = sign_bit(X);
    *pA = sign_bit(A);
    if (count < 5) {
	*pA |= magnitude(A) >> (6 * count);
	*pX |= CELL_MAX & (magnitude(X) >> (6 * count))
			& (A << (30 - 6 * count));
    } else if (count < 10)
	*pX |= magnitude(A) >> (6 * count - 30);
    else
	;
}

void shift_left(Cell A, Cell X, unsigned long count, Cell *pA, Cell *pX)
{
    *pX = sign_bit(X);
    *pA = sign_bit(A);
    if (count < 5) {
	*pX |= CELL_MAX & (X << (6 * count));
	*pA |= CELL_MAX & (A << (6 * count)) 
			& (magnitude(X) >> (30 - 6 * count));
    } else if (count < 10)
	*pA |= CELL_MAX & (X << (6 * count - 30));
    else
	;
}

/* Pre: count < 10 */
void shift_left_circular(Cell A, Cell X, unsigned count, Cell *pA, Cell *pX)
{
    Cell A1 = count < 5 ? A : X;
    Cell X1 = count < 5 ? X : A;
    unsigned c = 6 * (count < 5 ? count : count - 5);
    *pX = sign_bit(X)
	| (CELL_MAX & (X1 << c) & (magnitude(A1) >> (30 - c)));
    *pA = sign_bit(A)
	| (CELL_MAX & (A1 << c) & (magnitude(X1) >> (30 - c)));
}

/* --- Printable representation --- */

void print_cell(Cell cell)
{
    printf("%s%010lo", sign_bit(cell) == 0 ? " " : "-", magnitude(cell));
}

void unparse_cell(char *buffer, Cell cell)
{
    sprintf(buffer, "%s%010lo", sign_bit(cell) == 0 ? " " : "-", magnitude(cell));
}

/* --- Addresses --- */

Cell address2cell(Address addr) { return addr; }

Address cell2address(Cell cell)
{
    if ((sign_bit(cell) != 0 && magnitude(cell) != 0)
       || memory_size <= magnitude(cell)) {
	char buffer[12];
	unparse_cell(buffer, cell);
	error("Value is not an address: %s", buffer);
	return 0;
    }
    return (Address) magnitude(cell);
}
