/* MIX simulator, copyright 1994 by Darius Bacon */ 
#include "mix.h"
#include "charset.h"
#include "io.h"
#include "run.h"

#include <stdio.h>

/* --- Device tables --- */

/* Device types: */
enum DeviceType { tape, disk, card_in, card_out, printer, console };

/* The device table: */
static struct {
    const enum DeviceType type;
    FILE *file;
    const char *filename;
} device_table[] = {
    {tape}, {tape}, {tape}, {tape}, {tape}, {tape}, {tape}, {tape}, 
    {disk}, {disk}, {disk}, {disk}, {disk}, {disk}, {disk}, {disk}, 
    {card_in, stdin},
    {card_out, stdout},
    {printer, stdout},
    {console}
};

#define num_devices ( sizeof device_table / sizeof device_table[0] )

/* add an assign_file(device, filename) operation, too */
/* and unassign? */

typedef void IOHandler(unsigned, Cell, Address);
typedef void IOCHandler(unsigned, Cell);

static IOHandler tape_in, disk_in, text_in, console_in, no_in;
static IOHandler tape_out, disk_out, text_out, console_out, no_out;
static IOCHandler tape_ioc, disk_ioc, no_ioc, printer_ioc, console_ioc;

/* The device-class table: */
/*** need to distinguish read/write permission... */
static const struct {
    unsigned block_size;
    IOHandler *in_handler;
    IOHandler *out_handler;
    IOCHandler *ioc_handler;
} method_table[] = {

/* tape */      { 100, tape_in, tape_out,   tape_ioc },
/* disk */      { 100, disk_in, disk_out,   disk_ioc },
/* card_in */   {  16, text_in, no_out,     no_ioc },
/* card_out */  {  16, no_in,   text_out,   no_ioc },
/* printer */   {  24, no_in,   text_out,   printer_ioc },
/* console */   {  14, console_in, console_out, console_ioc }

};

static unsigned block_size(unsigned device)
{
    return method_table[device_table[device].type].block_size;
}

static FILE *assigned_file(unsigned device)
{
    return device_table[device].file;
}

void io_control(Byte device, Cell argument)
{
    unsigned d = (unsigned) device;
    if (num_devices <= d)
	error("Unknown device - %02o", d);
    method_table[device_table[d].type].ioc_handler(d, argument);
}

void do_input(Byte device, Cell argument, Address buffer)
{
    unsigned d = (unsigned) device;
    if (num_devices <= d)
	error("Unknown device - %02o", d);
    if (!assigned_file(d))
	error("No file assigned to device %02o", d);
    method_table[device_table[d].type].in_handler(d, argument, buffer);
}

void do_output(Byte device, Cell argument, Address buffer)
{
    unsigned d = (unsigned) device;
    if (num_devices <= d)
	error("Unknown device - %02o", d);
    if (!assigned_file(d))
	error("No file assigned to device %02o", d);
    method_table[device_table[d].type].out_handler(d, argument, buffer);
}

/* --- Unsupported input or output --- */

static void no_ioc(unsigned device, Cell argument)
{
    error("IOC undefined for device %02o", device);
}

static void no_in(unsigned device, Cell argument, Address buffer)
{
    error("Input not allowed for device %02o", device);
}

static void no_out(unsigned device, Cell argument, Address buffer)
{
    error("Output not allowed for device %02o", device);
}

/* --- Tapes --- */

static void tape_ioc(unsigned device, Cell offset)
{
    error("Unimplemented");
}

static void tape_in(unsigned device, Cell argument, Address buffer)
{
    error("Unimplemented");
}

static void tape_out(unsigned device, Cell argument, Address buffer)
{
    error("Unimplemented");
}

/* --- Disks --- */

static void disk_ioc(unsigned device, Cell offset)
{
    error("Unimplemented");
}

static void disk_in(unsigned device, Cell argument, Address buffer)
{
    error("Unimplemented");
}

static void disk_out(unsigned device, Cell argument, Address buffer)
{
    error("Unimplemented");
}

/* --- Text devices --- */

static void printer_ioc(unsigned device, Cell argument)
{
    if (magnitude(argument) != 0)
	error("IOC argument undefined for printer device %02o", device);
    if (!assigned_file(device))
	error("No stream assigned to printer device %02o", device);
    fputc('\f', assigned_file(device));
}

/* Read a line from -file- into memory[buffer..buffer+size). */
/* (Big-endian byte order, padded with 0 bytes if the line is less */
/* than -size- cells long.  The signs of the cells are set to '+'. */
/* If the line is longer than 5*size bytes, only the first 5*size  */
/* bytes get read. */
static void read_line(FILE *file, Address buffer, unsigned size)
{
    unsigned i, b;
    Flag past_end = false;
    for (i = 0; i < size; ++i) {
	Cell cell = zero;
	if (memory_size <= buffer+i)        /*** I think we need memory_fetch() and memory_store() functions... */
	    error("Address out of range");
	for (b = 1; b <= 5; ++b) {
	    Byte mix_char;
	    if (past_end)
		mix_char = (Byte) 0;
	    else {
		int c = fgetc(file);
		if (c == '\n' || c == EOF)
		    past_end = true, mix_char = (Byte) 0;
		else
		    mix_char = ascii2mix((char)c);
	    }
	    cell = set_byte(mix_char, b, cell);
	}
	memory[buffer+i] = cell;
    }
}

static void text_in(unsigned device, Cell argument, Address buffer)
{
    read_line(assigned_file(device), buffer, block_size(device));
}

static void write_cell(Cell cell, FILE *outfile)
{
    unsigned i;
    for (i = 1; i <= 5; ++i)
	fputc(mix2ascii(get_byte(i, cell)), outfile);
}

static void write_line(FILE *file, Address buffer, unsigned size)
{
    unsigned i;
    for (i = 0; i < size; ++i) {
	if (memory_size <= buffer+i)
	    error("Address out of range");
	write_cell(memory[buffer+i], file);
    }
    fputc('\n', file);
}

static void text_out(unsigned device, Cell argument, Address buffer)
{
    write_line(assigned_file(device), buffer, block_size(device));
}

/* --- The console (typewriter/paper tape) --- */
/* The console is a bit of a pain because it needs 2 open files to work. */
/* I'm not implementing it until I need to use it. */

static void console_ioc(unsigned device, Cell offset)
{
    error("Console device unimplemented");
}

static void console_in(unsigned device, Cell argument, Address buffer)
{
    error("Console device unimplemented");
}

static void console_out(unsigned device, Cell argument, Address buffer)
{
    error("Console device unimplemented");
}
